/*
 * gc3 - Gram's Commander v3.3: A configureable file manager
 *
 * YAGH! (tm) (Yet Another Gramish Hack)
 *
 * (c) 1994 by Graham Wheeler.           All Rights Reserved.
 *             gram@gramix.aztec.co.za
 *
 * DISCLAIMER: The author is not responsible for any loss, damage,
 *        riots, earthquakes, catastrophes, marriages, annulments,
 *        etc that arise from the use of this program. Sorry...
 *
 * This software may be freely copied and distributed, provided it is
 * distributed with all files intact and unmodified. You may use this
 * software for a 30 day trial period, after which you should register
 * your copy with me. I have spent a large number of weekends and nights
 * trying to make a file manager for UNIX (and DOS) which combines ease
 * of use and power, and complements the power of UNIX to process the
 * contents of files with the power to *select* the files to be processed.
 * By registering, you will encourage me to keep improving gc3, instead
 * of seeking paying work which benefits only a few.
 *
 * Full details of how to register, and what you will receive if
 * you do, are contained in the file LEGAL.DOC. Of course, if you
 * don't register, I can't do much about it, but in that case
 * may an obscure bug in gc3 erase all of your files!
 * 
 * Enjoy using GC!
 *
 * See INSTALL.DOC for instructions on how to install GC3
 *
 * Version History
 * ---------------
 * v1.0   Nov 1992         Initial release
 * v2.0   Nov 1992         Largely rewritten to be table-driven
 * v2.1   Dec 1992         Support for non-ANSI C compilers contributed
 *                           by George Sipe
 *                           Support for old ioctls TCGETA and TCSETAW for
 *                           those with no POSIX-style tcgetattr and tcsetattr.
 * v2.2   Dec 1992         Termio ioctls removed.Hopefully much more
 *                           portable. Command mode removed. A few bugs
 *                           fixed. Placeholder replacement now done in
 *                           a preprocessing phase before parsing commands.
 * v2.3   Jan 1993         Split into a number of source files
 *                           Basic port to MS-DOS
 * v3.0   Feb/March '93    New script language; large parts rewritten
 * v3.1   March '93        Ported back to UNIX and cleaned up
 * v3.2b  July  '93        Many bugs fixed and features added
 * v3.2   Oct   '93        Viewer added, .gc3rul file handling
 *                           added, bugs fixed, local variables and
 *                           parameters, tree view and container handling.
 * v3.3   April '94        Fixed lots of bugs. Large parts of compiler
 *                           and interpreter are now table-driven. Changed
 *                           screen repainting, and added support for colour
 *                           and mono terminals. Added script-driven menus
 *                           and forms, and hypertext help browser.
 *                           Made windows moveable and resizeable.
 *                           Made command handling in the script much more
 *                           sophisticated with multiple prefix/suffixes.
 *                           Improved command line history facility. Added
 *                           repeat counts to most commands.
 */


#include "gcport.h"
#define GC_CURSES
#include "gc3.h"
#include "gclib.h"
#include "gcops.h"
#include "gckey.h"
#include "gccc.h"

#ifdef DEBUG
FILE *debug=NULL;
#endif

extern int errno;

short *codespace;
char	*stringspace;

int	l;		/* Current active panel	*/
short	escapeCh;
int	numfiles[2]={0,0},
	fIndex[2][MAXARGS],
	highlight[2],
	startLn[2],
	selCnt[2],
	metaKey = 0,	/* Special key for keyboards without Alt- */
	writeIP,
	showIP = 0,
	forceCurrent = 0,
	metaMode = 0;
long	selSize[2];
short	info2show=1;

/*******************************************
	CODE DISASSEMBLER
********************************************/

static int symbolicDump = 1;

#if __STDC__
static void dumpstrings(void)
#else
static void dumpstrings()
#endif
{
	int i = 0;
	fprintf(stdout,"\n\nSTRING SPACE\n====== =====\n\n");
	while (i<stringp)
	{
		fprintf(stdout,"%5d: %s\n",i,LITERAL(i));
		i += strlen(LITERAL(i))+1;
	}
	fprintf(stdout,"\n\n");
}

#if __STDC__
static void dumpsyms(void)
#else
static void dumpsyms()
#endif
{
	int i;
	fprintf(stdout,"\n\nSYMBOL SPACE\n====== =====\n\n");
	for (i=0;i<idents;i++)
	{
		if (objects[i].type&FUNC_TYPE)
			fprintf(stdout,"%3d %15s function  %4d (params %d)\n",
				i,ID_NAME(i),
				(int)objects[i].p,
				(int)objects[i].np);
		else fprintf(stdout,"%3d %15s variable   %d\n",
			i,ID_NAME(i),
			(int)objects[i].p);
	}
}

#if __STDC__
static int getNextFunction(int p)
#else
static int getNextFunction(p)
	int p;
#endif
{
	int fp, rtn=0, ep=codep;
	for (fp=0;fp<idents;fp++)
	{
		if (objects[fp].type & FUNC_TYPE)
		{
			if ((objects[fp].p > p) && (objects[fp].p <ep ))
			{
				ep = objects[fp].p;
				rtn = fp;
			}
		}
	}
	return rtn;
}

#if __STDC__
void dumpcode(void)
#else
void dumpcode()
#endif
{
	short i = 1, j;
	int fp = getNextFunction(0);
	fprintf(stdout,"\n\nCODE SPACE\n====== =====\n\n");
	while (i<codep)
	{
		if (i==(int)objects[fp].p)
		{
			fprintf(stdout,"\n%s:\n",ID_NAME(fp));
			fp = getNextFunction(i);
		} else if (i==initialiseEntryPoint)
			fprintf(stdout,"\nGCInit:\n");
		else if (i==startupEntryPoint)
			fprintf(stdout,"\nGCStart:\n");
		else if (i==defaultEntryPoint)
			fprintf(stdout,"\nGCDefault:\n");
		if ((j = dumpOp(stdout,i,symbolicDump-1))==i) break;
		else i = j;
	}
	fprintf(stdout,"\nStartup entry point:    %d\n",startupEntryPoint);
	fprintf(stdout,"\nInitialise entry point: %d\n",initialiseEntryPoint);
	fprintf(stdout,"\nDefault entry point:    %d\n",defaultEntryPoint);
	fprintf(stdout,"\nCode size: %d\n",codep);
	fprintf(stdout,"\nString size: %d\n",stringp);
	fprintf(stdout,"\nSymbols: %d\n",idents);
}

/************************************************************
		SCRIPT COMPILER INTERFACE
*************************************************************/

int verbose = 0, dontLink = 0;

#if __STDC__
int Compile(char *script)
#else
int Compile(script)
	char *script;
#endif
{
	if (script == NULL) script = "gc3.key";
#ifdef __MSDOS__
	define("DOS");
#else
	define("UNIX");
#endif
	define("ANY");
#ifdef __MSDOS__
	if (findReadPath(script, "$GCPATH\\.$PATH", NULL, codepath))
#else
	if (findReadPath(script, "$GCPATH$HOME.$PATH/", NULL, codepath))
#endif
		keydeffile = fopen(codepath, "r");
	if (keydeffile) {
		fprintf(stderr,"Opened file: %s for reading\n", codepath);
		config_file();
		fclose(keydeffile);
		if (verbose)
			fprintf(stderr,"Compiled %d lines to %d bytes\n",lineno,codep);
		if (dontLink) ;
		else if (gclink()==0) {
			if (!initialiseEntryPoint) warn(ER_NOINIT);
			if (!defaultEntryPoint) warn(ER_NODEFAULT);
			if (!startupEntryPoint) warn(ER_NOSTART);
			writecodefile();
		} else error(ER_LINKER);
	} else error(ER_KEYFILE);
	if (verbose && !dontLink)
			fprintf(stderr,"After link: code %d strings %d\n",codep,stringp);
	fprintf(stderr,"Compilation successful!\n");
	if (symbolicDump)
	{
		lineno = -1;
		dumpstrings();
		dumpsyms();
		dumpcode();
	}
	freeStab();
	return 0;
}

/***********************************************************/

#if	__STDC__
static void showDefaultMsg(void)
#else
static void showDefaultMsg()
#endif /* __STDC__ */
{
	if (!showIP)
	{
		char *msg = valspace[VAR_MESSAGE];
		if (msg && msg[0]) writeMsg(msg);
	}
}

static int repeatCount = 0;

#if	__STDC__
static int nextKey(void)
#else
static int nextKey()
#endif /* __STDC__ */
{
   int key;
   static int lastKey;
   if (repeatCount>0)
   {
      repeatCount--;
      return lastKey;
   }
  	if (escapeCh != -1)  {
  		key = escapeCh;
  		escapeCh = -1;
  	} else key = my_getch();
  	if (key == metaKey) {
  		metaMode = 1;
		showMsg("<META>");
  		key = my_getch();
  	}
   if (key>='2' && key <='9' && getKeyBinding(key, 0)<0) /* repeat count */
   {
      repeatCount = key - '1';
  		lastKey = key = my_getch();
   }
   return key;
}

#if	__STDC__
static void loop(void)
#else
static void loop()
#endif /* __STDC__ */
{
	int bnd, key;
	winInvalid[0] = winInvalid[1] = 2;
	escapeCh = -1;
	metaMode = 0;
	showDefaultMsg();
	for (;;) {
		i_paint();
		showCursor(1-l,0);
		showCursor(l,1);
		i_sync();
#ifdef __MSDOS__
		locateCursor(25,1);
#endif
      key = nextKey();
		hideCursor(1-l); hideCursor(l);
		clearCmdWin();
		showDefaultMsg();
		bnd = getKeyBinding(key, 1);
		if (bnd<0) {
			if (isprint((int)key)) {
				if (defaultEntryPoint)
					if (execute(defaultEntryPoint,key,0)==1)
						break;
			} else BEEP;
		} else if (bnd==999) break;
		if (afterKeyEntryPoint)
			if (execute(afterKeyEntryPoint,key,0)==1)
				break;
	}
}

#if __STDC__
static void initTempDir(void)
#else
static void initTempDir()
#endif
{
	int b = getBuffer(2733);
	char *tn = BUF[b];
#ifdef __MSDOS__
	if (getenv("TMP"))
		strcpy(tn,getenv("TMP"));
	else if (getenv("GCPATH"))
		strcpy(tn,getenv("GCPATH"));
	else strcpy(tn,"c:\\");
	if ( tn[ strlen(tn)-1 ] != '\\')
		strcat(tn,"\\");
#endif
	strcat(tn, tmpnam(NULL));
	if ((valspace[VAR_TEMPPATH] = calloc(strlen(tn)+1,sizeof(char))) == NULL)
	{
		fprintf(stderr,"gc3 - FATAL MEMORY ALLOCATION ERROR!");
		exit(-1);
	}
	sprintf(tn,"mkdir %s", strcpy(valspace[VAR_TEMPPATH], tn));
	system(tn);
	/* check for success */
	if (checkType(getFileMode(tn+6),1)==0)
	{
		fprintf(stderr,"gc3 FATAL ERROR - Cannot make scratch directory %s!",tn+6);
		exit(-1);
	}
	releaseBuffer(b,7834);
}

#if __STDC__
static void removeTempDir(void)
#else
static void removeTempDir()
#endif
{
	if (valspace[VAR_TEMPPATH])
	{
		int b = getBuffer(2734);
#ifdef __MSDOS__
		sprintf(BUF[b],"deltree /Y %s",valspace[VAR_TEMPPATH]);
#else
		sprintf(BUF[b],"rm -r %s",valspace[VAR_TEMPPATH]);
#endif
		/*runCommand(BUF[b],0);*/
		system(BUF[b]);
		releaseBuffer(b,7835);
		free(valspace[VAR_TEMPPATH]);
		valspace[VAR_TEMPPATH] = NULL;
	}
}

#if	__STDC__
char *genTempName(void)
#else
char *genTempName()
#endif
{
	static char tn[MAXPATHNAME];
#ifdef __MSDOS__
	return strcpy(tn, makePath( valspace[VAR_TEMPPATH], tmpnam(NULL) ) );
#else
	char *p = STRRCHR(tmpnam(NULL), PATH_SEP)+1;
	return strcpy(tn, makePath( valspace[VAR_TEMPPATH], p) );
#endif
}

#if __STDC__
static void initCriticalVars(void)
#else
static void initCriticalVars()
#endif
{
	/* Always do the following few inits */
	assign2var(VAR_DISPLAYON,	"n");
	assign2var(VAR_VIEWER,		"n");
	assign2var(VAR_TREE,			"0");
	assign2var(VAR_EXPANDING,	"y");
	assign2var(VAR_DISPLAYON,	"y");
}

#if	__STDC__
static void initialise(char *p1, char *p2, int force)
#else
static void initialise(p1, p2, force)
	int force;
	char *p1, *p2;
#endif
{
	int j, tmp;
#ifdef DEBUG
	debug = FILEOPEN("gc.bug","w");
#endif
	sWinInit();
	fInfo[0] = (fInfo_t *)calloc(MAXARGS,sizeof(fInfo_t));
	fInfo[1] = (fInfo_t *)calloc(MAXARGS,sizeof(fInfo_t));
	if (fInfo[0]==NULL || fInfo[1]==NULL) {
		fprintf(stderr,"Cannot allocate memory for directory info!\n");
		exit(-1);
	}
	for (j=0;j<MAX_BUFFERS;j++) {
		if ((BUF[j] = (char *)calloc(BUFFER_SIZE,sizeof(char)))==NULL) {
			fprintf(stderr,"Failed to allocate temporary buffers!\n");
			exit(-1);
		}
	}
	initKeyTable();
	/* Set up very basic default screen attributes */
#ifdef __MSDOS__
	initscr(); /* Do this now as it just gets the screen type (color/mono) */
#endif
	setAttrEntry(0,-1,0);
	setAttrEntry(1,-1,2);
	setAttrEntry(2,-1,2);
	setAttrEntry(3,-1,0);
	tmp = readcode();
	/* readcode must must come before readIniFile,
		as readcode sets up initialiseEntryPoint */
	if (readIniFile()==0 && !force)
		initialiseEntryPoint = 0; /* Don't call gcinit */
	dirInit(p1,p2);
	if (tmp!=0) {
		int e = errno;
#ifdef NO_STRERROR
		fprintf(stderr,"Cannot open code file: errno %d\n",e);
#else
		fprintf(stderr,"Cannot open code file: %s\n",
						(char *)strerror(e));
#endif
		exit(0);
	}
	initTempDir();
	strcpy(tempName,genTempName());
	setupCurses(1);
#ifndef __MSDOS__
	if (getKeycode(CED_CLEAR) == 0)
		setKeycode(CED_CLEAR, (short)killchar());
	if (getKeycode(CED_BKSPC) == 0)
		setKeycode(CED_BKSPC, (short)erasechar());
#endif
	clearFilter(0);
	clearFilter(1);
	initCriticalVars();
	if (initialiseEntryPoint) {
		char *shell;
		/* Init option vars and call gcinit */
		assign2var(VAR_DIRSELECT,	"n");
		assign2var(VAR_DIRSFIRST,	"y");
		assign2var(VAR_SHOWHIDDEN,	"y");
		assign2var(VAR_FOLLOWLINKS,"y");
		assign2var(VAR_DEBUGGING,	"n");
		assign2var(VAR_NUMBERLNS,	"n");
		assign2var(VAR_ISCOLOR,	"y");
#ifdef __MSDOS__
		assign2var(VAR_USEDOSPATTERNS,	"y");
		if ((shell = getenv("COMSPEC"))==NULL)
			shell = "command";
#else
		assign2var(VAR_USEDOSPATTERNS,	"n");
		if ((shell = getenv("SHELL"))==NULL)
			shell = "sh";
#endif
		assign2var(VAR_SHELL,shell);
		if (initialiseEntryPoint)
			(void)execute(initialiseEntryPoint,0,0);
	}
}

#if __STDC__
void shutdown(void)
#else
void shutdown()
#endif
{
	int j;
	initCriticalVars();
	saveIniFile();
	if (catchFP)
		FILECLOSE(catchFP);
	unlink(tempName);
	removeTempDir();
	freeVars();
	freeRules();
	freeNames(0); free(fInfo[0]);
	freeNames(1); free(fInfo[1]);
	for (j=0;j<MAX_BUFFERS;j++) free(BUF[j]);
	sWinShutdown();
#ifdef DEBUG
	FILECLOSE(debug);
#endif
#ifdef __MSDOS__
	freeScreens();
#endif
	gc_check(); /* Show any files still open if FILE_DEBUG defined */
}

#if __STDC__
static void useage(void)
#else
static void useage()
#endif
{
	fprintf(stderr,"Useage: gc3 [flags] [<path1> [<path2>] ]\n");
	fprintf(stderr,"where the flags are:\n");
	fprintf(stderr,"\t-c0 supresses the use of colour.\n");
	fprintf(stderr,"\t-c1 enables the use of colour.\n");
	fprintf(stderr,"\t-C[<script>] forces a script recompile; the default script\n");
	fprintf(stderr,"\t\tname if one is not specified is `gc3.key'.\n");
	fprintf(stderr,"\t-d[n][l] writes a disassembled listing of the compiled script\n");
	fprintf(stderr,"\t\tto stdout. The `n' option produces a numeric dump, while\n");
	fprintf(stderr,"\t\tthe `l' option causes the link phase to be skipped.\n");
	fprintf(stderr,"\t-D<id> allows you to define identifiers for conditional\n");
	fprintf(stderr,"\t\tcompilation of the script (used with -C).\n");
	fprintf(stderr,"\t-i shows the current instruction pointer value at bottom.\n");
	fprintf(stderr,"\t-I forces the gcinit function to be called.\n");
	fprintf(stderr,"\t-t and -T write an instruction trace to gc3.ip, useful for\n");
	fprintf(stderr,"\t\tpost-mortem debugging. -T starts from gcinit\n");
	fprintf(stderr,"\t\twhile -t starts from the key processing loop.\n");
	fprintf(stderr,"\t-v causes the compiler to output more verbose messages.\n");
	exit(0);
}

#if	__STDC__
int main(int argc, char *argv[]) {
#else
int main(argc, argv)
	int argc;
	char **argv;
{
#endif /* __STDC__ */
	int i=1;
	char *script = NULL;
	short forceInit = 0, IPdump = 0, useColor=-1;
	writeIP = symbolicDump = forceCurrent = dontLink = 0;
	while (i<argc && argv[i][0]=='-') {
		switch(argv[i][1]) {
		case 'c':
			useColor = argv[i][2]-'0';
			if (useColor < 0 || useColor > 1) useage();
			break;
		case 'C':
			if (argv[i][2]) script = argv[i]+2;
			else script = "gc3.key";
			break;
		case 'd':
		{
			int j = 2;
			symbolicDump = 2;
			while (argv[i][j])
			{
				if (argv[i][j]=='n') symbolicDump = 1;
				else if (argv[i][j]=='l') dontLink = 1;
				j++;
			}
			break;
		}
		case 'D':
			STRUPR(argv[i]+2); define(argv[i]+2);
			if (script==NULL) script = "gc3.key"; /* Force a compile */
			break;
		case 'F': forceCurrent = 1;
			break;
		case 'I': forceInit = 1;
			break;
		case 'i': showIP = 1;
			break;
		case 'T': writeIP = 1;
		case 't': IPdump = 1;
#ifdef __MSDOS__
			unlink("\\gc3.ip");
#else
			unlink("/tmp/gc3.ip");
#endif
			break;
		case 'v': verbose = 1;
			break;
		default:
			useage();
		}
		i++;
	}
	stringspace = (char *)calloc(STRINGSPACE,sizeof(char));
	codespace = (short *)calloc(CODESPACE,sizeof(short));
	assert(codespace != NULL && stringspace != NULL);
	if (symbolicDump || script) 
	{
		Compile(script);
		/* A bit presumptious here... */
		HyperCompile("gc3help.txt");
	}
	else
	{
		initialise(argv[i], (i<argc)?argv[i+1]:NULL, forceInit);
		if (useColor==0) assign2var(VAR_ISCOLOR,"n");
		else if (useColor==1) assign2var(VAR_ISCOLOR,"y");
		if (startupEntryPoint) (void)execute(startupEntryPoint,0,0);
		writeIP=IPdump;
		loop();
		shutdown();
	}
	free(codespace);
	free(stringspace);
	return 0;
}



